from typing import Generic, Iterable, TypeVar

from rainbowadn.core.hash_point_format import hash_point_format, tabulate
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.core.recursivementionable import RecursiveMentionable
from rainbowadn.core.resolverorigin import ResolverOrigin

__all__ = ('Block', 'BlockFactory',)

HeaderType = TypeVar('HeaderType')
StateType = TypeVar('StateType')


class Block(RecursiveMentionable, Generic[HeaderType, StateType]):
    def __init__(
            self,
            previous: NullableReference['Block[HeaderType, StateType]'],
            header: HashPoint[HeaderType],
            state: HashPoint[StateType],
    ):
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(state, HashPoint)
        self.previous = previous
        self.header = header
        self.state = state

    def points(self) -> Iterable[HashPoint]:
        return [*self.previous.points(), self.header, self.state]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.header) + bytes(self.state)

    def __factory__(self) -> RainbowFactory['Block[HeaderType, StateType]']:
        return BlockFactory(
            self.header.factory,
            self.state.factory,
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'{await self.previous.str(tab)}' \
               f'{tabulate(tab)}(' \
               f'{tabulate(tab + 1)}block' \
               f'{tabulate(tab + 1)}(header)' \
               f'{tabulate(tab + 1)}{await hash_point_format(self.header, tab + 1)}' \
               f'{tabulate(tab + 1)}(state)' \
               f'{tabulate(tab + 1)}{await hash_point_format(self.state, tab + 1)}' \
               f'{tabulate(tab)})'


class BlockFactory(RainbowFactory[Block[HeaderType, StateType]], Generic[HeaderType, StateType]):
    def __init__(
            self,
            header_factory: RainbowFactory[HeaderType],
            state_factory: RainbowFactory[StateType],
    ):
        assert isinstance(header_factory, RainbowFactory)
        assert isinstance(state_factory, RainbowFactory)
        self.header_factory = header_factory
        self.state_factory = state_factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> Block[HeaderType, StateType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return Block(
            NullableReferenceFactory(self).from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            ResolverOrigin(
                self.header_factory, source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver
            ).hash_point(),
            ResolverOrigin(
                self.state_factory, source[2 * HashPoint.HASH_LENGTH:3 * HashPoint.HASH_LENGTH], resolver
            ).hash_point(),
        )
