from typing import Generic, Iterable, TypeVar

from rainbowadn.chain.reduction.reductionresult import ReductionResult
from rainbowadn.core.hash_point_format import hash_point_format, tabulate
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.core.recursivementionable import RecursiveMentionable
from rainbowadn.core.resolverorigin import ResolverOrigin

__all__ = ('Reduction', 'ReductionFactory',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class Reduction(
    ReductionResult[ReductorType, AccumulatorType],
    RecursiveMentionable,
    Generic[ReductorType, AccumulatorType]
):
    def __init__(self, reductor: HashPoint[ReductorType], accumulator: HashPoint[AccumulatorType]):
        assert isinstance(reductor, HashPoint)
        assert isinstance(accumulator, HashPoint)
        self.reductor = reductor
        self.accumulator = accumulator

    def points(self) -> Iterable[HashPoint]:
        return [self.reductor, self.accumulator]

    def __bytes__(self):
        return bytes(self.reductor) + bytes(self.accumulator)

    def __factory__(self) -> RainbowFactory['Reduction[ReductorType, AccumulatorType]']:
        return ReductionFactory(self.reductor.factory, self.accumulator.factory)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(reduction)' \
               f'{tabulate(tab)}{await hash_point_format(self.accumulator, tab)}'


class ReductionFactory(
    RainbowFactory[Reduction[ReductorType, AccumulatorType]],
    Generic[ReductorType, AccumulatorType]
):
    def __init__(
            self,
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
    ):
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        self.reductor_factory = reductor_factory
        self.accumulator_factory = accumulator_factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> Reduction[ReductorType, AccumulatorType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return Reduction(
            ResolverOrigin(self.reductor_factory, source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(self.accumulator_factory, source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )
