from typing import Generic, TypeVar

from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.trees.binary.actions.binaryaction import BinaryAction
from rainbowadn.data.collection.trees.binary.actions.symmetric import InnerOuter, OuterInner, Symmetric
from rainbowadn.data.collection.trees.binary.binarybalancing import BinaryBalancing
from rainbowadn.data.collection.trees.binary.binarycreation import BinaryCreation
from rainbowadn.data.collection.trees.binary.binaryprotocolized import BinaryProtocolized
from rainbowadn.data.collection.trees.binary.binarysplit import BinarySplit
from rainbowadn.data.collection.trees.binary.protocolizedbinarysplit import ProtocolizedBinarySplit

__all__ = ('AVL',)

ActiveKeyType = TypeVar('ActiveKeyType')
TreeType = TypeVar('TreeType')


class AVL(BinaryBalancing[ActiveKeyType, Integer, TreeType]):
    def empty_metadata(self) -> HashPoint[Integer]:
        return HashPoint.of(Integer(0))

    async def metadata(
            self,
            treel: TreeType,
            treer: TreeType,
            key: HashPoint[ActiveKeyType],
            protocol: BinaryCreation[ActiveKeyType, Integer, TreeType]
    ) -> HashPoint[Integer]:
        assert isinstance(protocol, BinaryCreation)
        return HashPoint.of(
            Integer(
                1
                +
                max(
                    await self.height(BinaryProtocolized(protocol, treel)),
                    await self.height(BinaryProtocolized(protocol, treer))
                )
            )
        )

    @classmethod
    async def height(
            cls,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> int:
        return await HeightAction().on(protocolized)

    async def balance(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> TreeType:
        return await BalanceAction().on(protocolized)


class HeightAction(
    BinaryAction[ActiveKeyType, Integer, TreeType, int],
    Generic[ActiveKeyType, TreeType]
):
    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> int:
        return 0

    async def on_split(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, Integer, TreeType]
    ) -> int:
        metadata: Integer = await case.split.metadata.resolve()
        return metadata.integer


class BalanceAction(
    BinaryAction[ActiveKeyType, Integer, TreeType, TreeType],
    Generic[ActiveKeyType, TreeType]
):
    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> TreeType:
        return protocolized.tree

    async def on_split(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        delta = (await AVL.height(case.protocolizedl())) - (await AVL.height(case.protocolizedr()))
        assert isinstance(delta, int)
        if delta < -1:
            return await self.on_symmetric(InnerOuter(case.protocol), case.split)
        elif delta > 1:
            return await self.on_symmetric(OuterInner(case.protocol), case.split)
        else:
            return case.tree

    @classmethod
    async def on_symmetric(
            cls,
            symmetric: Symmetric[ActiveKeyType, Integer, TreeType],
            split: BinarySplit[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        assert isinstance(symmetric, Symmetric)
        assert isinstance(split, BinarySplit)
        splito = await symmetric.protocol.fsplit(symmetric.outer(split))
        if (
                (await AVL.height(symmetric.protocolizedi(splito)))
                >
                (await AVL.height(symmetric.protocolizedo(splito)))
        ):
            splitoi = await symmetric.protocol.fsplit(symmetric.inner(splito))
            return await symmetric.tree(
                await symmetric.tree(symmetric.inner(split), symmetric.inner(splitoi), split.key),
                await symmetric.tree(symmetric.outer(splitoi), symmetric.outer(splito), splito.key),
                splitoi.key
            )
        else:
            return await symmetric.tree(
                await symmetric.tree(symmetric.inner(split), symmetric.inner(splito), split.key),
                symmetric.outer(splito),
                splito.key
            )
