from typing import AsyncIterable, Iterable

from rainbowadn.core.asserts import assert_false, assert_true
from rainbowadn.core.hash_point_format import hash_point_format, tabulate
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.nullability.null import Null
from rainbowadn.core.nullability.nullable import Nullable
from rainbowadn.core.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.core.recursivementionable import RecursiveMentionable
from rainbowadn.core.resolverorigin import ResolverOrigin
from rainbowadn.core.static import StaticFactory, StaticMentionable
from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.keymetadata import KeyMetadata, KeyMetadataFactory
from rainbowadn.data.collection.trees.binary.activebinarytree import ActiveBinaryTree
from rainbowadn.data.collection.trees.binary.avl import AVL
from rainbowadn.data.collection.trees.binary.binarytree import BinaryTree, BinaryTreeFactory
from rainbowadn.data.collection.trees.comparison.comparator import Fail
from rainbowadn.data.collection.trees.comparison.hashcomparator import HashComparator
from rainbowadn.v13.algo import MINT_CONST
from rainbowadn.v13.subject import Subject
from rainbowadn.v13.transaction import Coin, Transaction

__all__ = ('BankState',)


class BankState(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            minted: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            used: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            miner: NullableReference[Subject],
            length: HashPoint[Integer]
    ):
        assert isinstance(miner, NullableReference)
        assert isinstance(used, NullableReference)
        assert isinstance(miner, NullableReference)
        assert isinstance(length, HashPoint)
        self.minted = minted
        self.used = used
        self.miner = miner
        self.length = length

    def points(self) -> Iterable[HashPoint]:
        return [*self.minted.points(), *self.used.points(), *self.miner.points(), self.length]

    def __bytes__(self):
        return bytes(self.minted) + bytes(self.used) + bytes(self.miner) + bytes(self.length)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'BankState':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        reference_factory: RainbowFactory[
            NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]]
        ] = NullableReferenceFactory(
            BinaryTreeFactory(
                KeyMetadataFactory(
                    StaticFactory(Coin),
                    StaticFactory(Integer),
                ).loose()
            ).loose()
        ).loose()
        assert isinstance(reference_factory, RainbowFactory)
        return cls(
            reference_factory.from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            reference_factory.from_bytes(source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver),
            NullableReferenceFactory(
                StaticFactory(Subject)
            ).from_bytes(source[2 * HashPoint.HASH_LENGTH:3 * HashPoint.HASH_LENGTH], resolver),
            ResolverOrigin(Integer.factory(), source[3 * HashPoint.HASH_LENGTH:], resolver).hash_point()
        )

    def without_miner(self) -> 'BankState':
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            self.length
        )

    async def advance(self) -> 'BankState':
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            HashPoint.of(Integer((await self.length.resolve()).integer + 1))
        )

    def minted_tree(self) -> ActiveBinaryTree[Coin, Integer]:
        return ActiveBinaryTree(
            AVL(HashComparator(Fail())), self.minted
        )

    def used_tree(self) -> ActiveBinaryTree[Coin, Integer]:
        return ActiveBinaryTree(
            AVL(HashComparator(Fail())), self.used
        )

    async def use_coins(self, coins: AsyncIterable[HashPoint[Coin]]) -> 'BankState':
        minted: ActiveBinaryTree[Coin, Integer] = self.minted_tree()
        used: ActiveBinaryTree[Coin, Integer] = self.used_tree()
        async for in_coin in coins:
            assert_true(await minted.contains(in_coin))
            assert_false(await used.contains(in_coin))
            used = await used.add(in_coin)
        return BankState(self.minted, used.reference, self.miner, self.length)

    async def mint_coins(
            self,
            transaction: Transaction
    ) -> 'BankState':
        assert_true(await self.verify(transaction))
        miner = self.miner_nullable()
        minted: ActiveBinaryTree[Coin, Integer] = self.minted_tree()
        async for coin, miner in transaction.iter_coins(self.mint(), miner):
            assert_false(await minted.contains(HashPoint.of(coin)))
            minted = await minted.add(HashPoint.of(coin))
            assert isinstance(minted, ActiveBinaryTree)
        return BankState(minted.reference, self.used, NullableReference(miner, self.miner.factory), self.length)

    def miner_nullable(self) -> Nullable[HashPoint[Subject]]:
        return self.miner.reference

    def mint(self) -> int:
        if self.miner_nullable().null():
            return MINT_CONST
        else:
            return 0

    async def verify(self, transaction: Transaction) -> bool:
        assert isinstance(transaction, Transaction)
        assert_true(await transaction.verify(self.mint()))
        return True

    async def _push(self, transaction: Transaction) -> 'BankState':
        assert isinstance(transaction, Transaction)
        return await (
            await self.use_coins(
                (await transaction.data_resolved()).iter_in_coins()
            )
        ).mint_coins(
            transaction
        )

    async def push(self, transaction: HashPoint[Transaction]) -> 'BankState':
        return await self._push(await transaction.resolve())

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(' \
               f'{tabulate(tab + 1)}bank' \
               f'{tabulate(tab + 1)}(miner)' \
               f'{tabulate(tab + 1)}{await self.miner.str(tab + 1)}' \
               f'{tabulate(tab + 1)}(minted)' \
               f'{tabulate(tab + 1)}{await self.minted.str(tab + 1)}' \
               f'{tabulate(tab + 1)}(used)' \
               f'{tabulate(tab + 1)}{await self.used.str(tab + 1)}' \
               f'{tabulate(tab + 1)}(length)' \
               f'{tabulate(tab + 1)}{await hash_point_format(self.length, tab + 1)}' \
               f'{tabulate(tab)})'
