from typing import AsyncIterable, Iterable

import nacl.signing

from rainbowadn.core.asserts import assert_true
from rainbowadn.core.hash_point_format import hash_point_format, tabulate
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.nullability.notnull import NotNull
from rainbowadn.core.nullability.nullable import Nullable
from rainbowadn.core.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.core.recursivementionable import RecursiveMentionable
from rainbowadn.core.resolverorigin import ResolverOrigin
from rainbowadn.core.static import StaticMentionable
from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.stack.stack import Stack, StackFactory
from rainbowadn.v13.signature import Signature
from rainbowadn.v13.subject import Subject

__all__ = ('CoinData', 'Coin', 'TransactionData', 'Transaction',)


class CoinData(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            owner: HashPoint[Subject],
            value: HashPoint[Integer]
    ):
        assert isinstance(owner, HashPoint)
        assert isinstance(value, HashPoint)
        self.owner = owner
        self.value = value

    async def int_value(self) -> int:
        return (await self.value.resolve()).integer

    @classmethod
    def of(cls, owner: Subject, value: int) -> 'CoinData':
        assert isinstance(owner, Subject)
        assert isinstance(value, int)
        return cls(HashPoint.of(owner), HashPoint.of(Integer(value)))

    def points(self) -> Iterable[HashPoint]:
        return [self.owner, self.value]

    def __bytes__(self):
        return bytes(self.owner) + bytes(self.value)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'CoinData':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return cls(
            ResolverOrigin(Subject.factory(), source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(Integer.factory(), source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'{await hash_point_format(self.owner, tab)}' \
               f'{tabulate(tab)}{await hash_point_format(self.value, tab)}'


class Coin(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            data: HashPoint[CoinData],
            origin: HashPoint['Transaction'],
            index: HashPoint[Integer]
    ):
        assert isinstance(data, HashPoint)
        assert isinstance(origin, HashPoint)
        assert isinstance(index, HashPoint)
        self.data = data
        self.origin = origin
        self.index = index

    async def data_resolved(self) -> CoinData:
        return await self.data.resolve()

    def points(self) -> Iterable[HashPoint]:
        return [self.data, self.origin, self.index]

    def __bytes__(self):
        return bytes(self.data) + bytes(self.origin) + bytes(self.index)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'Coin':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return cls(
            ResolverOrigin(CoinData.factory(), source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(
                Transaction.factory(), source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver
            ).hash_point(),
            ResolverOrigin(Integer.factory(), source[2 * HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(' \
               f'{tabulate(tab + 1)}coin' \
               f'{tabulate(tab + 1)}{await hash_point_format(self.data, tab + 1)}' \
               f'{tabulate(tab + 1)}(origin)' \
               f'{tabulate(tab + 1)}{await hash_point_format(self.index, tab + 1)}' \
               f'{tabulate(tab)})'


class TransactionData(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            in_coins: NullableReference[Stack[Coin]],
            out_coins: NullableReference[Stack[CoinData]],
    ):
        assert isinstance(in_coins, NullableReference)
        assert isinstance(out_coins, NullableReference)
        self.in_coins = in_coins
        self.out_coins = out_coins
        self.hash_point = HashPoint.of(self)
        assert isinstance(self.hash_point, HashPoint)

    def points(self) -> Iterable[HashPoint]:
        return [*self.in_coins.points(), *self.out_coins.points()]

    def __bytes__(self):
        return bytes(self.in_coins) + bytes(self.out_coins)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'TransactionData':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return cls(
            NullableReferenceFactory(
                StackFactory(Coin.factory()).loose()
            ).from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            NullableReferenceFactory(
                StackFactory(CoinData.factory()).loose()
            ).from_bytes(source[HashPoint.HASH_LENGTH:], resolver),
        )

    async def _verify_signatures(
            self,
            signatures: NullableReference[Stack[Signature]]
    ) -> bool:
        for coin, signature in zip(
                [x async for x in self.iter_in_coins_resolved()],
                [x async for x in Stack.iter(signatures)],
                strict=True
        ):
            assert_true(
                (await signature.resolve()).verify(
                    await (await coin.data_resolved()).owner.resolve(),
                    self.hash_point
                )
            )
        return True

    def iter_in_coins(self) -> AsyncIterable[HashPoint[Coin]]:
        return Stack.iter(self.in_coins)

    async def iter_in_coins_resolved(self) -> AsyncIterable[Coin]:
        async for coin in self.iter_in_coins():
            yield await coin.resolve()

    async def _total_in(self) -> int:
        return sum([await (await coin.data_resolved()).int_value() async for coin in self.iter_in_coins_resolved()])

    def iter_out_coins(self) -> AsyncIterable[HashPoint[CoinData]]:
        return Stack.iter(self.out_coins)

    async def _total_out(self) -> int:
        return sum([await (await coin.resolve()).int_value() async for coin in self.iter_out_coins()])

    async def _verify_values(self, mint: int) -> bool:
        assert isinstance(mint, int)
        assert (await self.extra(mint)) >= 0
        return True

    async def extra(self, mint: int) -> int:
        assert isinstance(mint, int)
        return (await self._total_in()) + mint - (await self._total_out())

    async def verify(
            self,
            signatures: NullableReference[Stack[Signature]],
            mint: int
    ) -> bool:
        assert isinstance(signatures, NullableReference)
        assert isinstance(mint, int)
        assert_true(await self._verify_signatures(signatures))
        assert_true(await self._verify_values(mint))
        return True

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(in)' \
               f'{tabulate(tab)}{await self.in_coins.str(tab)}' \
               f'{tabulate(tab)}(out)' \
               f'{tabulate(tab)}{await self.out_coins.str(tab)}'


class Transaction(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            data: HashPoint[TransactionData],
            signatures: NullableReference[Stack[Signature]]
    ):
        assert isinstance(data, HashPoint)
        assert isinstance(signatures, NullableReference)
        self.data = data
        self.signatures = signatures
        self.hash_point = HashPoint.of(self)
        assert isinstance(self.hash_point, HashPoint)

    async def data_resolved(self) -> TransactionData:
        return await self.data.resolve()

    def points(self) -> Iterable[HashPoint]:
        return [self.data, *self.signatures.points()]

    def __bytes__(self):
        return bytes(self.data) + bytes(self.signatures)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'Transaction':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        signature_factory: RainbowFactory[Signature] = Signature.factory()
        assert isinstance(signature_factory, RainbowFactory)
        stack_factory: RainbowFactory[Stack[Signature]] = StackFactory(signature_factory).loose()
        assert isinstance(stack_factory, RainbowFactory)
        return cls(
            ResolverOrigin(TransactionData.factory(), source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            NullableReferenceFactory(stack_factory).from_bytes(source[HashPoint.HASH_LENGTH:], resolver),
        )

    async def iter_coins(
            self,
            mint: int,
            miner: Nullable[HashPoint[Subject]]
    ) -> AsyncIterable[tuple[Coin, Nullable[HashPoint[Subject]]]]:
        transaction_data: TransactionData = await self.data_resolved()
        assert isinstance(transaction_data, TransactionData)
        index = 0
        out_coin: HashPoint[CoinData]
        async for out_coin in transaction_data.iter_out_coins():
            assert isinstance(out_coin, HashPoint)
            if miner.null():
                miner = NotNull((await out_coin.resolve()).owner)
                assert isinstance(miner, Nullable)
            coin: Coin = Coin(out_coin, self.hash_point, HashPoint.of(Integer(index)))
            assert isinstance(coin, Coin)
            yield coin, miner
            index += 1
        if not miner.null():
            coin: Coin = Coin(
                HashPoint.of(
                    CoinData(
                        miner.resolve(),
                        HashPoint.of(Integer(await transaction_data.extra(mint)))
                    )
                ),
                self.hash_point,
                HashPoint.of(Integer(index))
            )
            assert isinstance(coin, Coin)
            yield coin, miner

    async def coins(
            self,
            mint: int,
            miner: Nullable[HashPoint[Subject]]
    ) -> list[Coin]:
        assert isinstance(mint, int)
        assert isinstance(miner, Nullable)
        return [coin async for coin, _ in self.iter_coins(mint, miner)]

    async def verify(self, mint: int):
        assert isinstance(mint, int)
        data: TransactionData = await self.data_resolved()
        assert isinstance(data, TransactionData)
        assert_true(await data.verify(self.signatures, mint))
        return True

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(' \
               f'{tabulate(tab + 1)}transaction' \
               f'{tabulate(tab + 1)}{hash_point_format(self.data, tab + 1)}' \
               f'{tabulate(tab + 1)}{self.signatures.str(tab + 1)}' \
               f'{tabulate(tab)})'

    @classmethod
    def make(
            cls,
            in_coins: list[Coin],
            out_coins: list[CoinData],
            keys: list[nacl.signing.SigningKey],
    ) -> 'Transaction':
        transaction_data = TransactionData(
            Stack.off(Coin.factory(), reversed(in_coins)),
            Stack.off(CoinData.factory(), reversed(out_coins)),
        )
        assert isinstance(transaction_data, TransactionData)
        return Transaction(
            HashPoint.of(transaction_data),
            Stack.off(
                Signature.factory(),
                (Signature.sign(key, HashPoint.of(transaction_data)) for key in reversed(keys))
            )
        )
