from typing import Iterable

from rainbowadn.core.asserts import assert_eq
from rainbowadn.core.hash_point_format import hash_point_format, tabulate
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.localmetaorigin import LocalMetaOrigin
from rainbowadn.core.localorigin import LocalOrigin
from rainbowadn.core.mentionable import Mentionable
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.core.recursivementionable import RecursiveMentionable
from rainbowadn.core.resolverorigin import ResolverOrigin
from rainbowadn.data.atomic.integer import Integer
from rainbowadn.wrisbt.weakreferenceindexsetbtree import WeakReferenceIndexSetBTree, WrisbtFactory
from rainbowadn.wrisbt.wrisbtparametres import WrisbtParametres

__all__ = ('WrisbtRoot', 'WrisbtRootFactory',)


class WrisbtRoot(RecursiveMentionable):
    def __init__(self, root: HashPoint[WeakReferenceIndexSetBTree], height: int, parametres: WrisbtParametres):
        assert isinstance(root, HashPoint)
        assert isinstance(height, int)
        assert isinstance(parametres, WrisbtParametres)
        self.root = root
        self.height = height
        self.parametres = parametres

    async def root_resolved(self) -> 'WeakReferenceIndexSetBTree':
        return await self.root.resolve()

    def points(self) -> Iterable[HashPoint]:
        return [self.root]

    def __bytes__(self):
        return bytes(self.root) + bytes(Integer(self.height))

    def __factory__(self) -> RainbowFactory['WrisbtRoot']:
        return WrisbtRootFactory(self.parametres)

    @classmethod
    def empty(cls, parametres: WrisbtParametres) -> 'WrisbtRoot':
        assert isinstance(parametres, WrisbtParametres)
        return WrisbtRoot(HashPoint.of(WeakReferenceIndexSetBTree(b'', 0, parametres, True, ())), 0, parametres)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(root)' \
               f'{tabulate(tab)}{self.height}' \
               f'{tabulate(tab)}{await hash_point_format(self.root, tab)}'

    async def contains(self, key: bytes) -> bool:
        assert isinstance(key, bytes)
        assert_eq(len(key), self.parametres.keysize)

        root: WeakReferenceIndexSetBTree = await self.root_resolved()
        assert isinstance(root, WeakReferenceIndexSetBTree)

        return await root.contains(key)

    async def add(self, key: bytes) -> 'WrisbtRoot':
        assert isinstance(key, bytes)
        assert_eq(len(key), self.parametres.keysize)

        root: WeakReferenceIndexSetBTree = await (await self.root_resolved()).add(key)
        assert isinstance(root, WeakReferenceIndexSetBTree)

        if root.full():
            left, middle, right = root.split()
            assert isinstance(left, WeakReferenceIndexSetBTree)
            assert isinstance(middle, bytes)
            assert isinstance(right, WeakReferenceIndexSetBTree)
            root = WeakReferenceIndexSetBTree(
                middle + bytes(HashPoint.of(left)) + bytes(HashPoint.of(right)),
                root.height + 1,
                root.parametres,
                True,
                (LocalMetaOrigin(LocalOrigin(left)), LocalMetaOrigin(LocalOrigin(right)))
            )
            assert isinstance(root, WeakReferenceIndexSetBTree)
        return self.of(root)

    @classmethod
    def of(cls, root: WeakReferenceIndexSetBTree) -> 'WrisbtRoot':
        assert isinstance(root, WeakReferenceIndexSetBTree)
        return cls(HashPoint.of(root), root.height, root.parametres)

    async def keys(self) -> list[bytes]:
        return list([x async for x in (await self.root_resolved()).iter_keys()])

    async def index(
            self, target: HashPoint, exclude: 'WrisbtRoot'
    ) -> 'WrisbtRoot':
        assert isinstance(target, HashPoint)
        assert isinstance(exclude, WrisbtRoot)
        key: bytes = target.point
        assert isinstance(key, bytes)
        if await exclude.contains(key) or await self.contains(key):
            return self
        tree: WrisbtRoot = self
        value: Mentionable = await target.resolve()
        assert isinstance(value, Mentionable)
        if isinstance(value, RecursiveMentionable):
            for hash_point in value.points():
                tree = await tree.index(hash_point, exclude)
        tree = await tree.add(key)
        return tree


class WrisbtRootFactory(RainbowFactory[WrisbtRoot]):
    def __init__(self, parametres: WrisbtParametres):
        assert isinstance(parametres, WrisbtParametres)
        self.parametres = parametres

    def from_bytes(self, source: bytes, resolver: HashResolver) -> WrisbtRoot:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        height: int = Integer.from_bytes(source[HashPoint.HASH_LENGTH:], resolver).integer
        assert isinstance(height, int)
        return WrisbtRoot(
            ResolverOrigin(
                WrisbtFactory(height, self.parametres, True), source[:HashPoint.HASH_LENGTH], resolver
            ).hash_point(),
            height,
            self.parametres
        )
